unit filebu;
(**********************************************************************************
*****                                                                         *****
*****     FileBu                                                              *****
*****     -----------------------------------------------------------------   *****
*****     One procedure can be used to create a chain of 'gapped' backups.    *****
*****                                                                         *****
*****     CHAIN                                                               *****
*****     Today's backup replaces yesterdays which replaces the day before's  *****
*****                                                                         *****
*****     GAPPED                                                              *****
*****     Enforce a minimum time between overwriting backups.  This means:    *****
*****       o  Preventing the chain being swamped by many frequent backups    *****
*****       o  Having a short chain to efficiently keep quite old copies      *****
*****     -----------------------------------------------------------------   *****
*****     BackupChain is the main function and is the one to call from        *****
*****     outside.  Documentation is provided in the source                   *****
*****     -----------------------------------------------------------------   *****
*****     History                                                             *****
*****     24 Jan 00  First version                                            *****
*****     31 Jan 00  Level numbers have leading zeroes rather than spaces     *****
*****     -----------------------------------------------------------------   *****
*****     Copyright Peter Fox 2000           peter@eminent.demon.co.uk        *****
***********************************************************************************)


interface

uses  dialogs,useful,sysutils;

  type TbafResult = (bafrOK,bafrNoSource,bafrUnknownError,bafrNoNeed);

  //'PUBLIC'
  function BackupChain(const Source,DestinationMask:string; gaps:array of double):boolean;
  procedure TouchFile(const aFileName:string; const stamp:tdatetime);

  //'PRIVATE'
  function AgeDifference(const NewerFile,OlderFile:string):double;
  function IndexedFileName(const deststem:string; const ix:integer):string;
  procedure dobackupchain(const source,deststem:string; gaps:array of double; const level:integer);

implementation

function BackupChain(const Source,DestinationMask:string; gaps:array of double):boolean;
{ This function does the whole business of maintaing a series of backups of SOURCE file.
  SOURCE ........... Full name of file to be backed up
  DESTINATIONMASK .. String specifying the format of the backup files
                     eg "\safety\<path><name>.B<index>" with a source of '\ut\afile.txt' would generate
                     files of '\safety\ut\afile.B01, ...B02, ...B03 etc
                     Tokens:  NB CASE SENSITIVE
                              <drive> .. Directory in source [extractFileDrive()]
                              <path> ... Path in the source  [extractfilepath()]
                              <name> ... Name in the source  [extractfilename()]
                              <ext> .... Extension in the source [extractfileext()]
                              <index> .. Two digit counter 01,02,03 etc
                              NB THE <index> TOKEN IS MANDATORY
  GAPS ............. Minimum interval between successive backups in days or fractions
                     (This can be cast from Tdatetime if you want)
                     [0] ... age difference required to update backup 01
                     [1] ...  "       "         "     "   "       "   02
                     [2...n]  as long a chain as you want.
                     For example suppose you want set up a daily,weekly,monthly chain
                     then you would call using [1,7,30]

  The result is false if an unresovable problem is encountered.

  NOTE:  Backups are arranged in a chain.  When we are looking to overwrite the
  backup at the top of the chain we see if it is sufficiently older than the next
  in the chain.  If it is then backup this first.  If not then this will get
  overwritten.
      Example:    File to backup is fresh
                  Backup-01  3  days ago
                  Backup-02  12 days ago
                  Backup-03  33 days ago
      If we call today with gaps of 1,7,30 then:-
          01 is 3 days older than current so overwrite it...
          ...but before we do possibly bump 01 down the chain...
          02 is 9 days older than 01 so overwrite it...
          ...but before we do possibly bump 02 down the chain...
          03 is 21 days older than 02 which is not enough gap so:
            Overwrite 02 with 01
            Overwrite 01 with current.

  Warning : The file that you're backing up might not change its date when
  whatever other activities your application carries out are done.  Therefore
  you need either to force a backup each time with gaps of zero or reset the
  file date (use touchfile procedure) after each session.

  For a backup every day you don't really want to have gaps[0] as 1.0.  Why?
  Because If on Monday you started work at 09:05 and on Tuesday at 08:55 you
  would'nt get a backup because on Tuesday you're starting 23 hours and 50 minutes
  after the previous session.  Therefor you might set gaps[0] as 0.75 to be sure
  of getting the overwrite.

  If you want to backup every time to at least the top of the chain then
  gaps[0]=0 will do the trick.

}
var
  deststem : string;
  sdrive,spath,sname,sext:string;
begin
  // get the bits of the file we need for the destination mask
  result := FullparseOfFileName(source,sdrive,spath,sname,sext);
  if not result then exit;                                      // silly filename

  // the source does exist - doesn't it?
  result := fileexists(source);
  if not result then exit;                                      // non-existent source

  // build the destination filename excluding the index
  result := (pos('<index>',DestinationMask)<>0);
  if not result then exit;                                      // no index token silly
  deststem := DestinationMask;
  deststem := substr(deststem,'<drive>',sdrive);
  deststem := substr(deststem,'<path>',spath);
  deststem := substr(deststem,'<name>',sname);
  deststem := substr(deststem,'<ext>',sext);

  dobackupchain(source,deststem,gaps,0);
end;

procedure dobackupchain(const source,deststem:string; gaps:array of double; const level:integer);
{ backup this level of the chain (level 0 = source tries to o/w backup-01)}
var
  dfile : string;
begin
  if level > high(gaps) then exit;   // fallen off end of chain!
  dfile := IndexedFileName(deststem,level+1);
  if AgeDifference(source,dfile)<gaps[level] then exit;  // source is not significantly older - nothing to do
  dobackupchain(dfile,deststem,gaps,level+1);           // possibly backup the destination first
  if fileexists(dfile) then deletefile(dfile);          // belt-and-braces
  if level = 0 then begin            // top level so copy which sets file date
    ufilecopy(source,dfile);
  end else begin
    renamefile(source,dfile);        // renaming keeps the timestamp
  end;
end;

function IndexedFileName(const deststem:string; const ix:integer):string;
{ return filename substituting "<index>" in deststem }
begin
  result := format(substr(deststem,'<index>','%2.2d'),[ix]);
end;

function AgeDifference(const NewerFile,OlderFile:string):double;
{ what is the difference in ages of these files in days
  If the either file doesn't exist then report 100 years
  It can return a negative value.
  If the result is +ve then it is can be cast to TdateTime
   }
begin
  // one or both of the files doesn't exist - return 100 years
  if not (fileexists(NewerFile) and fileexists(olderfile)) then begin
    result := 35624;
    exit;
  end;
  result := double(FileDateToDateTime(fileage(NewerFile)))-
            double(FileDateToDateTime(fileage(OlderFile)));
end;

procedure TouchFile(const aFileName:string; const stamp:tdatetime);
{ set file's date stamp }
var
  FHandle: integer;
  i : integer;
begin
  FHandle := FileOpen(aFileName,fmOpenWrite);
  if fhandle > 0 then begin
    i := fileSetDate(Fhandle,DateTimeToFileDate(stamp));  // ignore any errors by throwing away result
    if i <>0 then beep;
  end else begin
    beep;
  end;    
  FileClose(FHandle);
end;


end.
